#include <adios.h>
#include <adios_read.h>

int adiosReader_1D (char*, size_t, int8_t**, int16_t**, int32_t**, int64_t**, uint8_t**, uint16_t**, uint32_t**, uint64_t**, float**, double**);

int adiosReader_2D (char*, size_t, size_t, int8_t**, int16_t**, int32_t**, int64_t**, uint8_t**, uint16_t**, uint32_t**, uint64_t**, float**, double**);

int adiosReader_3D (char*, size_t, size_t, size_t, int8_t**, int16_t**, int32_t**, int64_t**, uint8_t**, uint16_t**, uint32_t**, uint64_t**, float**, double**);

int adiosReader_4D (char*, size_t, size_t, size_t, size_t, int8_t**, int16_t**, int32_t**, int64_t**, uint8_t**, uint16_t**, uint32_t**, uint64_t**, float**, double**);

int adiosReader_5D (char*, size_t, size_t, size_t, size_t, size_t, int8_t**, int16_t**, int32_t**, int64_t**, uint8_t**, uint16_t**, uint32_t**, uint64_t**, float**, double**);

