#ifndef H5Z_ZFP_VERSION_H
#define H5Z_ZFP_VERSION_H

#define H5Z_FILTER_ZFP 32013

#define H5Z_FILTER_ZFP_VERSION_MAJOR 1
#define H5Z_FILTER_ZFP_VERSION_MINOR 1
#define H5Z_FILTER_ZFP_VERSION_PATCH 0

#define H5Z_ZFP_MODE_RATE      1
#define H5Z_ZFP_MODE_PRECISION 2
#define H5Z_ZFP_MODE_ACCURACY  3
#define H5Z_ZFP_MODE_EXPERT    4
#define H5Z_ZFP_MODE_REVERSIBLE 5

#define H5Z_ZFP_CD_NELMTS_MEM 6
#define H5Z_ZFP_CD_NELMTS_MAX 6

#endif
